extends KinematicBody2D
class_name Player

const mass = 60.0
const max_radius = 80.0

var PickableObject = preload("res://src/Objects/PickableObject.tscn")
var Spritesheet = preload("res://assets/player_spritesheet.png")
var SpritesheetHolding = preload("res://assets/player_spritesheet-holding.png")
var SpritesheetOverhead = preload("res://assets/player_spritesheet-overhead.png")
var SpritesheetHoldingOverhead = preload("res://assets/player_spritesheet-holding-overhead.png")

var speed: = 60
var acceleration: = 12.0
var velocity: = Vector2.ZERO
var in_action: = false
var in_crafting: = false
var in_fishing: = false
var is_holding: = false
var shallow_action: = false
var sleeping: = false
var sleeping_late: = false
var indoors: = false
var on_boat: = false
var on_sand: = false
var on_wood: = false
var on_bridge: = false
var on_stone: = false
var stamina_locked: = false

onready  var aPlayer:AnimationPlayer = $AnimationPlayer
onready  var aTree:AnimationTree = $AnimationTree
onready  var aTreeState:AnimationNodeStateMachinePlayback = aTree.get("parameters/playback")
onready  var joystick: = $Joystick
onready  var player_sprite: = $Sprite
onready  var overhead_sprite: = $Cosmetics / OverheadSprite
onready  var stats: = $Stats as Stats
onready  var tools: = $ToolMenu as ToolMenu
onready  var inventory: = $Inventory as Inventory
onready  var interaction: = $InteractionArea as PlayerInteractionArea

signal hoe_ground(pos)
signal swing_axe(pos)
signal swing_pickaxe(pos)
signal water_ground(pos)
signal item_drag(item)
signal item_use(item, pos)
signal item_preview(item)
signal enter_sleep()
signal exit_sleep()
signal enter_indoors()
signal exit_outdoors()
signal enter_hold()
signal exit_hold()
signal snapshot_loaded()
signal tool_changed(tool_id)
signal water_level_changed(level)

func _ready()->void :
	GameUtils.camera = $Camera

	WorldState.player = self

	aTree.active = true

	update_blend_position(Vector2.DOWN)

	if ProjectSettings.get("custom/debug/free_movement"):
		$Collision.disabled = true
		$Camera.zoom = Vector2(0.5, 0.5)

func connect_stats(node):
	$Stats.connect("coins_changed", node, "_on_PlayerStats_coins_changed")

	connect("tool_changed", node, "_on_Player_tool_changed")
	connect("water_level_changed", node, "_on_Player_water_level_changed")

func enter_sleeping():
	sleeping = true

	if is_holding:
		exit_holding()

	GameUtils.close_all_windows()

	player_sprite.hide()
	overhead_sprite.hide()

	$SleepingSprite.show()
	$EmotesController.sleeping()

	if in_fishing:
		$Addons / FishingRod.hide()
		$Addons / FishingRod.pulled()

		_action_finished("Fishing")

	aTreeState.travel("Idle")

	emit_signal("enter_sleep")

func exit_sleeping():
	sleeping = false
	sleeping_late = false

	player_sprite.show()
	overhead_sprite.show()

	$SleepingSprite.hide()
	$EmotesController.ui_hide()

	emit_signal("exit_sleep")

func fall_asleep():
	sleeping_late = true

	enter_sleeping()

func enter_crafting(type = 1):
	enter_action()
	update_blend_position(Vector2.UP)

	in_crafting = true

	aTreeState.travel("Craft%s" % type)

	$CraftingSound.pitch_scale = rand_range(0.8, 1.2)

func exit_crafting():
	in_action = false
	in_crafting = false

	aTreeState.travel("Idle")

func enter_holding():
	$Sprite.texture = SpritesheetHolding
	$Cosmetics / OverheadSprite.texture = SpritesheetHoldingOverhead

	is_holding = true

	emit_signal("enter_hold")

func exit_holding(notify = true):
	$Sprite.texture = Spritesheet
	$Cosmetics / OverheadSprite.texture = SpritesheetOverhead

	is_holding = false
	
	if notify:
		emit_signal("item_preview", null)

	emit_signal("exit_hold")

func enter_action():
	in_action = true

func exit_action():
	in_action = false
	in_crafting = false

	if $Stats.stamina <= 0:
		$StaminaController.cancel_shake()

	if in_fishing:
		in_fishing = false

		$Addons / FishingRod.hide()
		$Addons / FishingRod.pulled()

		_action_finished("Fishing")

	$ThrowBobberTimer.stop()

	$ImprovedAction.active = false

	aTreeState.travel("Idle")

func stop_moving():
	velocity = Vector2.ZERO

func exit_action_and_stop():
	exit_action()
	stop_moving()

func rested(multiplier):
	$Stats.stamina = $Stats.max_stamina * multiplier

func exclamation():
	$EmotesController.exclamation()

func item_get(item_id:String, count = 1):
	var item = ItemsDb.by_id(item_id)
	
	if item:
		enter_action()

		$Addons / ItemShowcase / ItemSprite.frame = item.frame

		travel_state("ItemGot")

		if not $Inventory.add_item(item_id, count):
			var container = GameUtils.get_group_node("LevelItemsContainer")
			var produce_object = PickableObject.instance()

			produce_object.dropped_vertical = true
			produce_object.item_id = item_id
			produce_object.pickup_disabled_time = 10.0
			produce_object.global_position = global_position + Vector2(
				rand_range( - 5, 5), 
				rand_range( - 5, 5)
			)

			if container:
				container.add_child(produce_object)

func get_walk_direction():
	return aTree.get("parameters/Walk/blend_position")

func update_fish_showcase(fish_id:String):
	var fish = ItemsDb.by_id(fish_id)

	$Addons / ItemShowcase / ItemSprite.frame = fish.frame

func put_fish_to_inventory(fish_id:String):
	if not $Inventory.add_item(fish_id, 1):
		var container = GameUtils.get_group_node("LevelItemsContainer")
		var produce_object = PickableObject.instance()

		produce_object.dropped_vertical = true
		produce_object.item_id = fish_id
		produce_object.pickup_disabled_time = 10.0
		produce_object.global_position = global_position + Vector2(
			rand_range( - 5, 5), 
			rand_range( - 5, 5)
		)

		if container:
			container.add_child(produce_object)

func has_travel_supplies()->bool:
	return $Inventory.get_item_count(ItemsDb.travel_supplies_id) > 0

func use_travel_supply()->bool:
	return $Inventory.remove_item(ItemsDb.travel_supplies_id, 1)

func _physics_process(delta:float)->void :
	overhead_sprite.frame = player_sprite.frame

	if in_action or sleeping or Storyline.cutscene:
		return 
	
	if GameUtils.menu_active == "Dialog" or Storyline.rescuing:
		aTreeState.travel("Idle")

		return 

	var input_vector = Vector2(
		Input.get_action_strength("ui_dright") - Input.get_action_strength("ui_dleft"), 
		Input.get_action_strength("ui_ddown") - Input.get_action_strength("ui_dup")
	)

	if input_vector == Vector2.ZERO and not GameUtils.menu_active:
		input_vector = Vector2(
			Input.get_action_strength("ui_right") - Input.get_action_strength("ui_left"), 
			Input.get_action_strength("ui_down") - Input.get_action_strength("ui_up")
		)

	var joystick_direction = joystick.get_direction()
	var move_direction = input_vector.normalized()
	var move_speed = speed
	
	if joystick_direction != Vector2.ZERO:
		move_direction = joystick_direction

	if move_direction == Vector2.ZERO:
		aTreeState.travel("Idle")
	else :
		update_blend_position(move_direction)

		if (Input.is_action_pressed("ui_shift") or joystick.is_run_pressed()):
			if $Stats.stamina:
				if aTreeState.get_current_node() != "Run":
					$DashSound.play()

					$SprintDrainTimer.start()

					DustEmitter.emit_dust(self, - move_direction)

				aTreeState.travel("Run")

				move_speed = speed * 1.5
			else :
				$StaminaController.shake()

				aTreeState.travel("Walk")
		else :
			aTreeState.travel("Walk")

	velocity = lerp(velocity, move_direction * move_speed, delta * acceleration)

	velocity = move_and_slide(velocity)

func _unhandled_input(_event):
	if Storyline.cutscene:
		return 

	if is_holding and (Input.is_action_just_pressed("ui_interact") or Input.is_action_just_pressed("ui_click")):
		$HoldingPreview.use_current_item()

func _input(event):
	if sleeping or Storyline.cutscene:
		return 

	
	
	if not GameUtils.menu_active and event is InputEventScreenTouch:
		$Addons / Targeting.update_target_cell()

	if not GameUtils.menu_active and Input.is_action_just_pressed("ui_improved_action"):
		if is_holding:
			exit_holding()

			return 

		if in_fishing or in_action or is_holding:
			return 

		if not $Joystick.is_interact_pressed():
			var mouse_pos = get_global_mouse_position()

			if global_position.distance_to(mouse_pos) <= max_radius:
				update_blend_position(global_position.direction_to(mouse_pos))

		$ImprovedAction.active = true

		_perform_action()

	if not GameUtils.menu_active and (Input.is_action_just_pressed("ui_interact") or Input.is_action_just_pressed("ui_click")):
		if in_fishing:
			$Addons / FishingRod.pull()

			return 

		if in_action or is_holding:
			return 

		if Input.is_action_just_pressed("ui_click") and not $Joystick.is_interact_pressed():
			var mouse_pos = get_global_mouse_position()

			if global_position.distance_to(mouse_pos) > max_radius:
				return 

			update_blend_position(global_position.direction_to(mouse_pos))

		if $ToolMenu.get_water_level() <= 60 and $ToolMenu.active_tool != ToolMenu.Tools.FISHING_ROD and $FacingChecker.is_colliding() and $FacingChecker.get_collider().get_groups().has("OceanSurface"):
			$InteractionArea.refill_water(global_position + $FacingChecker.cast_to)
		else :
			var obj = WorldState.get_object(GameUtils.target_cell)

			if obj:
				if obj.interactable:
					if obj.interactable_by_hand_only and $ToolMenu.active_tool != - 1:
						_perform_action()

						return 

					
					if $ToolMenu.active_tool == ToolMenu.Tools.FISHING_ROD:
						if global_position.distance_to(obj.global_position) <= GameUtils.TILE_SIZE:
							obj.emit_signal("interact")
						else :
							_perform_action()
					else :
						obj.emit_signal("interact")
				elif obj.rotateable and $ToolMenu.active_tool < 0:
					obj.emit_signal("rotate_object")
				else :
					$ToolMenu.guess_tool(obj)

					_perform_action()
			else :
				_perform_action()

	if not GameUtils.menu_active and Input.is_action_just_pressed("ui_shoulder_accept"):
		if $ToolMenu.active_tool != ToolMenu.Tools.FISHING_ROD:
			return 

		if $Inventory.active or $ToolMenu.active or is_holding:
			return 

		if in_fishing:
			$Addons / FishingRod.pull()

			return 

		if not in_action:
			_perform_action()

	if Input.is_action_just_pressed("ui_inventory") and $ToggleUiTimer.is_stopped():
		if in_fishing:
			return 

		if $ToolMenu.active:
			$ToolMenu.hide_menu()

		$Inventory.ui_toggle()
		$ToggleUiTimer.start()

	if Input.is_action_just_pressed("ui_select_tool") and $Inventory.active and $ToggleUiTimer.is_stopped():
		if in_fishing:
			return 

		$Inventory.ui_hide()
		$ToolMenu.show_menu()
		$ToggleUiTimer.start()

	if Input.is_action_just_pressed("ui_cancel"):
		$Inventory.ui_hide()

func _perform_action():
	if GameUtils.menu_active or $ToolMenu.active_tool < 0 or $InteractionArea.connected_inventory:
		return 

	if $Stats.stamina > 0:
		enter_action()
	else :
		$StaminaController.shake()

		return 

	velocity = Vector2.ZERO

	match $ToolMenu.active_tool:
		ToolMenu.Tools.HOE:
			aTreeState.travel("Hoe")
		ToolMenu.Tools.AXE:
			aTreeState.travel("Axe")
		ToolMenu.Tools.PICKAXE:
			aTreeState.travel("Pickaxe")
		ToolMenu.Tools.WATERING_CAN:
			if $ToolMenu.get_water_level():
				$WaterFromCan.modulate = Color.white
			else :
				$WaterFromCan.modulate = Color.transparent

			aTreeState.travel("Watering")
		ToolMenu.Tools.FISHING_ROD:
			if not $ThrowBobberTimer.is_stopped():
				return 

			if not indoors and $Addons / FishingRod.throw_bobber(global_position, get_walk_direction()):
				GameUtils.close_all_windows()

				aTreeState.travel("FishingThrow")
			else :
				aTreeState.travel("FishingThrowError")

			$ThrowBobberTimer.start()

func _action_start(description:String):
	if shallow_action or not in_action:
		return 

	if description == "Watering":
		if $ToolMenu.get_water_level():
			SFXPlayer.water()
		else :
			SFXPlayer.error()

func _action_mid(description:String):
	if shallow_action:
		return 

	var tool_tile = GameUtils.target_cell
	var obj = WorldState.get_object(tool_tile)

	if description == "Hoe":
		SFXPlayer.ground_hit()

		emit_signal("hoe_ground", tool_tile)

		if obj:
			obj.emit_signal("player_hoe_ground")

		$ImprovedAction.hoe(tool_tile, get_walk_direction())

	if description == "Axe":
		SFXPlayer.swoosh()

		emit_signal("swing_axe", tool_tile)

		if obj:
			obj.emit_signal("player_swing_axe")

	if description == "Pickaxe":
		SFXPlayer.swoosh()

		emit_signal("swing_pickaxe", tool_tile)

		if obj:
			obj.emit_signal("player_swing_pickaxe")

	if description == "Watering":
		if not $ToolMenu.get_water_level():
			return 

		emit_signal("water_ground", tool_tile)

		if obj:
			obj.emit_signal("player_water_ground")

		$ImprovedAction.water(tool_tile, get_walk_direction())

		$ToolMenu.spend_water(6.0)

func _action_finished(_description:String):
	in_action = false
	in_crafting = false
	in_fishing = false
	shallow_action = false

	if Input.is_action_pressed("ui_improved_action"):
		_perform_action()

		return 
	else :
		$ImprovedAction.active = false

	
	if tools.active_tool == ToolMenu.Tools.FISHING_ROD:
		return 

	if Input.is_action_pressed("ui_click") or Input.is_action_pressed("ui_interact"):
		if Input.is_action_pressed("ui_click") and not $Joystick.is_interact_pressed():
			var mouse_pos = get_global_mouse_position()

			if global_position.distance_to(mouse_pos) > max_radius:
				return 

			update_blend_position(global_position.direction_to(mouse_pos))

		if Input.is_action_pressed("ui_interact"):
			var input_vector = Vector2(
				Input.get_action_strength("ui_dright") - Input.get_action_strength("ui_dleft"), 
				Input.get_action_strength("ui_ddown") - Input.get_action_strength("ui_dup")
			)

			if input_vector == Vector2.ZERO and not GameUtils.menu_active:
				input_vector = Vector2(
					Input.get_action_strength("ui_right") - Input.get_action_strength("ui_left"), 
					Input.get_action_strength("ui_down") - Input.get_action_strength("ui_up")
				)

			var joystick_direction = joystick.get_direction()
			var move_direction = input_vector.normalized()

			if joystick_direction != Vector2.ZERO:
				move_direction = joystick_direction

			if move_direction != Vector2.ZERO:
				update_blend_position(move_direction)

		_perform_action()

func _action_fishing():
	in_fishing = true

func is_facing_target(target:Vector2):
	var dir = get_walk_direction()
	var target_dir = global_position.direction_to(target)

	return abs(dir.angle_to(target_dir)) < 0.8

func update_blend_position(vector:Vector2)->void :
	aTree.set("parameters/Idle/blend_position", vector)
	aTree.set("parameters/Walk/blend_position", vector)
	aTree.set("parameters/Run/blend_position", vector)
	aTree.set("parameters/Hoe/blend_position", vector)
	aTree.set("parameters/Axe/blend_position", vector)
	aTree.set("parameters/Pickaxe/blend_position", vector)
	aTree.set("parameters/Watering/blend_position", vector)
	aTree.set("parameters/FishingThrow/blend_position", vector)
	aTree.set("parameters/FishingThrowError/blend_position", vector)
	aTree.set("parameters/FishingIdle/blend_position", vector)
	aTree.set("parameters/FishingPullEmpty/blend_position", vector)
	aTree.set("parameters/FishingPullWin/blend_position", vector)

	$FacingChecker.cast_to = vector * (GameUtils.TILE_SIZE * 0.75)

func travel_state(name:String)->void :
	aTreeState.travel(name)

func _on_InteractionArea_show_inventory():
	
	$Inventory.get_node("InteractCooldown").stop()

	$Inventory.ui_show()

func _on_InteractionArea_close_inventory():
	$Inventory.ui_hide()

func _on_MonitoringArea_item_picked_up(id, count):
	$Inventory.add_item(id, count)

func _on_Inventory_item_drag(item):
	var interaction_range = $InteractionArea / CollisionShape2D.shape.radius

	if position.distance_to(get_global_mouse_position()) <= interaction_range:
		emit_signal("item_drag", item)

func _on_PlantsManager_item_used(item):
	$Inventory.remove_item(item.id, 1)

func _on_InteractionArea_refill_water(from_pos:Vector2):
	enter_action()

	shallow_action = true

	update_blend_position(global_position.direction_to(from_pos))

	$WaterFromCan.modulate = Color.transparent

	aTreeState.travel("Watering")

func _on_InteractionArea_use_item(item:Dictionary):
	var dir = get_walk_direction()

	emit_signal("item_use", item, global_position + dir * 10.0)

func persist_save()->Dictionary:
	return {
		"filename":get_filename(), 
		"parent":get_parent().get_path(), 
		"path":get_path(), 
		"global_position":global_position, 

		"health":$Stats.health, 
		"stamina":$Stats.stamina, 
		"coins":$Stats.coins, 
		"max_health":$Stats.max_health, 
		"max_stamina":$Stats.max_stamina, 

		"hat_id":$Cosmetics.hat_id, 
		"pet_id":$PetController.pet_id, 

		"water_level":$ToolMenu.get_water_level(), 
		"tools_disabled":$ToolMenu.disabled
	}

func persist_load(data:Dictionary)->void :
	$Stats.max_health = data.max_health
	$Stats.max_stamina = data.max_stamina

	$Stats.health = data.health
	$Stats.stamina = data.stamina
	$Stats.coins = 99999999

	$ToolMenu.set_water_level(clamp(data.water_level, 0.0, 100.0))

	$ToolMenu.disabled = GameUtils.convert_float_array(data.tools_disabled)

	if data.has("hat_id"):
		$Cosmetics.update_hat(data.hat_id)

	if data.has("pet_id"):
		$PetController.pet_id = data.pet_id

func _on_PlayerHouse_player_indoors():
	indoors = true

	emit_signal("enter_indoors")

func _on_PlayerHouse_player_outdoors():
	indoors = false

	emit_signal("exit_outdoors")

func _on_SurfaceCheckArea_body_entered(body):
	if body.get_groups().has("SandSurface"):
		on_sand = true

	if body.get_groups().has("WoodSurface"):
		on_wood = true

	if body.get_groups().has("BridgeSurface"):
		on_bridge = true

	if body.get_groups().has("StoneSurface"):
		on_stone = true

func _on_SurfaceCheckArea_body_exited(body):
	if body.get_groups().has("SandSurface"):
		on_sand = false

	if body.get_groups().has("WoodSurface"):
		on_wood = false

	if body.get_groups().has("BridgeSurface"):
		on_bridge = false

	if body.get_groups().has("StoneSurface"):
		on_stone = false

func _on_Inventory_cell_consume(item):
	if item.info.energy:
		$Stats.stamina += item.info.energy

func _on_Inventory_cell_focus_changed(item):
	emit_signal("item_preview", item)

func _on_Inventory_cell_drop(item_id):
	var container = GameUtils.get_group_node("LevelItemsContainer")
	var produce_object = PickableObject.instance()
	var pivot = GameUtils.target_cell * GameUtils.TILE_SIZE + GameUtils.HALF_TILE

	produce_object.dropped_vertical = true
	produce_object.item_id = item_id
	produce_object.pickup_disabled_time = 10.0
	produce_object.global_position = pivot + Vector2(
		rand_range( - 5, 5), 
		rand_range( - 5, 5)
	)

	if container:
		container.add_child(produce_object)

func _on_SprintDrainTimer_timeout():
	$Stats.stamina -= GameUtils.RUNNING_STAMINA_COST

	if aTreeState.get_current_node() == "Run":
		$SprintDrainTimer.start()

func _on_ToolMenu_tool_changed(tool_id):
	emit_signal("tool_changed", tool_id)

func _on_ToolMenu_water_level_changed(level):
	emit_signal("water_level_changed", level)

func _on_LevelGenerator_end_loading():
	$PetController.spawn_pet()
